@@ -9,11 +9,11 @@
  * and implementing search-path-controlled searches.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/catalog/namespace.c,v 1.58.2.1 2007/04/20 02:38:46 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/catalog/namespace.c,v 1.73 2004/12/31 21:59:38 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -39,6 +39,7 @@
 #include "utils/acl.h"
 #include "utils/builtins.h"
 #include "utils/catcache.h"
+#include "utils/guc.h"
 #include "utils/inval.h"
 #include "utils/lsyscache.h"
 #include "utils/memutils.h"
@@ -64,32 +65,14 @@
  * SQL99.  Also, this provides a way to search the system namespace first
  * without thereby making it the default creation target namespace.)
  *
- * For security reasons, searches using the search path will ignore the temp
- * namespace when searching for any object type other than relations and
- * types.  (We must allow types since temp tables have rowtypes.)
- *
  * The default creation target namespace is normally equal to the first
  * element of the explicit list, but is the "special" namespace when one
  * has been set.  If the explicit list is empty and there is no special
  * namespace, there is no default target.
  *
- * The textual specification of search_path can include "$user" to refer to
- * the namespace named the same as the current user, if any.  (This is just
- * ignored if there is no such namespace.)  Also, it can include "pg_temp"
- * to refer to the current backend's temp namespace.  This is usually also
- * ignorable if the temp namespace hasn't been set up, but there's a special
- * case: if "pg_temp" appears first then it should be the default creation
- * target.  We kluge this case a little bit so that the temp namespace isn't
- * set up until the first attempt to create something in it.  (The reason for
- * klugery is that we can't create the temp namespace outside a transaction,
- * but initial GUC processing of search_path happens outside a transaction.)
- * tempCreationPending is TRUE if "pg_temp" appears first in the string but
- * is not reflected in defaultCreationNamespace because the namespace isn't
- * set up yet.
- *
- * In bootstrap mode, the search path is set equal to "pg_catalog", so that
+ * In bootstrap mode, the search path is set equal to 'pg_catalog', so that
  * the system namespace is the only one searched or inserted into.
- * The initdb script is also careful to set search_path to "pg_catalog" for
+ * The initdb script is also careful to set search_path to 'pg_catalog' for
  * its post-bootstrap standalone backend runs.	Otherwise the default search
  * path is determined by GUC.  The factory default path contains the PUBLIC
  * namespace (if it exists), preceded by the user's personal namespace
@@ -117,22 +100,24 @@ static Oid	defaultCreationNamespace = InvalidOid;
 /* first explicit member of list; usually same as defaultCreationNamespace */
 static Oid	firstExplicitNamespace = InvalidOid;
 
-/* if TRUE, defaultCreationNamespace is wrong, it should be temp namespace */
-static bool tempCreationPending = false;
-
-/* The above five values are valid only if namespaceSearchPathValid */
+/* The above four values are valid only if namespaceSearchPathValid */
 static bool namespaceSearchPathValid = true;
 
 /*
  * myTempNamespace is InvalidOid until and unless a TEMP namespace is set up
  * in a particular backend session (this happens when a CREATE TEMP TABLE
  * command is first executed).	Thereafter it's the OID of the temp namespace.
- * firstTempTransaction flags whether we've committed creation of the TEMP
- * namespace or not.
+ *
+ * myTempNamespaceSubID shows whether we've created the TEMP namespace in the
+ * current subtransaction.  The flag propagates up the subtransaction tree,
+ * so the main transaction will correctly recognize the flag if all
+ * intermediate subtransactions commit.  When it is InvalidSubTransactionId,
+ * we either haven't made the TEMP namespace yet, or have successfully
+ * committed its creation, depending on whether myTempNamespace is valid.
  */
 static Oid	myTempNamespace = InvalidOid;
 
-static bool firstTempTransaction = false;
+static SubTransactionId myTempNamespaceSubID = InvalidSubTransactionId;
 
 /*
  * "Special" namespace for CREATE SCHEMA.  If set, it's the first search
@@ -151,7 +136,7 @@ char	   *namespace_search_path = NULL;
 static void recomputeNamespacePath(void);
 static void InitTempTableNamespace(void);
 static void RemoveTempRelations(Oid tempNamespaceId);
-static void RemoveTempRelationsCallback(void);
+static void RemoveTempRelationsCallback(int code, Datum arg);
 static void NamespaceCallback(Datum arg, Oid relid);
 
 /* These don't really need to appear in any header file */
@@ -185,7 +170,9 @@ RangeVarGetRelid(const RangeVar *relation, bool failOK)
 		if (strcmp(relation->catalogname, get_database_name(MyDatabaseId)) != 0)
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-			   errmsg("cross-database references are not implemented")));
+					 errmsg("cross-database references are not implemented: \"%s.%s.%s\"",
+							relation->catalogname, relation->schemaname,
+							relation->relname)));
 	}
 
 	if (relation->schemaname)
@@ -238,7 +225,9 @@ RangeVarGetCreationNamespace(const RangeVar *newRelation)
 		if (strcmp(newRelation->catalogname, get_database_name(MyDatabaseId)) != 0)
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-			   errmsg("cross-database references are not implemented")));
+					 errmsg("cross-database references are not implemented: \"%s.%s.%s\"",
+					   newRelation->catalogname, newRelation->schemaname,
+							newRelation->relname)));
 	}
 
 	if (newRelation->istemp)
@@ -247,7 +236,7 @@ RangeVarGetCreationNamespace(const RangeVar *newRelation)
 		if (newRelation->schemaname)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_TABLE_DEFINITION),
-				   errmsg("temporary tables may not specify a schema name")));
+			  errmsg("temporary tables may not specify a schema name")));
 		/* Initialize temp namespace if first time through */
 		if (!OidIsValid(myTempNamespace))
 			InitTempTableNamespace();
@@ -256,14 +245,6 @@ RangeVarGetCreationNamespace(const RangeVar *newRelation)
 
 	if (newRelation->schemaname)
 	{
-		/* check for pg_temp alias */
-		if (strcmp(newRelation->schemaname, "pg_temp") == 0)
-		{
-			/* Initialize temp namespace if first time through */
-			if (!OidIsValid(myTempNamespace))
-				InitTempTableNamespace();
-			return myTempNamespace;
-		}
 		/* use exact schema given */
 		namespaceId = GetSysCacheOid(NAMESPACENAME,
 								CStringGetDatum(newRelation->schemaname),
@@ -279,12 +260,6 @@ RangeVarGetCreationNamespace(const RangeVar *newRelation)
 	{
 		/* use the default creation namespace */
 		recomputeNamespacePath();
-		if (tempCreationPending)
-		{
-			/* Need to initialize temp namespace */
-			InitTempTableNamespace();
-			return myTempNamespace;
-		}
 		namespaceId = defaultCreationNamespace;
 		if (!OidIsValid(namespaceId))
 			ereport(ERROR,
@@ -306,13 +281,13 @@ Oid
 RelnameGetRelid(const char *relname)
 {
 	Oid			relid;
-	List	   *lptr;
+	ListCell   *l;
 
 	recomputeNamespacePath();
 
-	foreach(lptr, namespaceSearchPath)
+	foreach(l, namespaceSearchPath)
 	{
-		Oid			namespaceId = lfirsto(lptr);
+		Oid			namespaceId = lfirst_oid(l);
 
 		relid = get_relname_relid(relname, namespaceId);
 		if (OidIsValid(relid))
@@ -350,11 +325,11 @@ RelationIsVisible(Oid relid)
 	/*
 	 * Quick check: if it ain't in the path at all, it ain't visible.
 	 * Items in the system namespace are surely in the path and so we
-	 * needn't even do oidMember() for them.
+	 * needn't even do list_member_oid() for them.
 	 */
 	relnamespace = relform->relnamespace;
 	if (relnamespace != PG_CATALOG_NAMESPACE &&
-		!oidMember(relnamespace, namespaceSearchPath))
+		!list_member_oid(namespaceSearchPath, relnamespace))
 		visible = false;
 	else
 	{
@@ -386,13 +361,13 @@ Oid
 TypenameGetTypid(const char *typname)
 {
 	Oid			typid;
-	List	   *lptr;
+	ListCell   *l;
 
 	recomputeNamespacePath();
 
-	foreach(lptr, namespaceSearchPath)
+	foreach(l, namespaceSearchPath)
 	{
-		Oid			namespaceId = lfirsto(lptr);
+		Oid			namespaceId = lfirst_oid(l);
 
 		typid = GetSysCacheOid(TYPENAMENSP,
 							   PointerGetDatum(typname),
@@ -432,11 +407,11 @@ TypeIsVisible(Oid typid)
 	/*
 	 * Quick check: if it ain't in the path at all, it ain't visible.
 	 * Items in the system namespace are surely in the path and so we
-	 * needn't even do oidMember() for them.
+	 * needn't even do list_member_oid() for them.
 	 */
 	typnamespace = typform->typnamespace;
 	if (typnamespace != PG_CATALOG_NAMESPACE &&
-		!oidMember(typnamespace, namespaceSearchPath))
+		!list_member_oid(namespaceSearchPath, typnamespace))
 		visible = false;
 	else
 	{
@@ -525,20 +500,16 @@ FuncnameGetCandidates(List *names, int nargs)
 		}
 		else
 		{
-			/*
-			 * Consider only procs that are in the search path and are not
-			 * in the temp namespace.
-			 */
-			List	   *nsp;
+			/* Consider only procs that are in the search path */
+			ListCell   *nsp;
 
 			foreach(nsp, namespaceSearchPath)
 			{
-				if (procform->pronamespace == lfirsto(nsp) &&
-					procform->pronamespace != myTempNamespace)
+				if (procform->pronamespace == lfirst_oid(nsp))
 					break;
 				pathpos++;
 			}
-			if (nsp == NIL)
+			if (nsp == NULL)
 				continue;		/* proc is not in search path */
 
 			/*
@@ -637,11 +608,11 @@ FunctionIsVisible(Oid funcid)
 	/*
 	 * Quick check: if it ain't in the path at all, it ain't visible.
 	 * Items in the system namespace are surely in the path and so we
-	 * needn't even do oidMember() for them.
+	 * needn't even do list_member_oid() for them.
 	 */
 	pronamespace = procform->pronamespace;
 	if (pronamespace != PG_CATALOG_NAMESPACE &&
-		!oidMember(pronamespace, namespaceSearchPath))
+		!list_member_oid(namespaceSearchPath, pronamespace))
 		visible = false;
 	else
 	{
@@ -657,7 +628,7 @@ FunctionIsVisible(Oid funcid)
 
 		visible = false;
 
-		clist = FuncnameGetCandidates(makeList1(makeString(proname)), nargs);
+		clist = FuncnameGetCandidates(list_make1(makeString(proname)), nargs);
 
 		for (; clist; clist = clist->next)
 		{
@@ -698,6 +669,8 @@ FuncCandidateList
 OpernameGetCandidates(List *names, char oprkind)
 {
 	FuncCandidateList resultList = NULL;
+	char	   *resultSpace = NULL;
+	int			nextResult = 0;
 	char	   *schemaname;
 	char	   *opername;
 	Oid			namespaceId;
@@ -724,6 +697,21 @@ OpernameGetCandidates(List *names, char oprkind)
 								 CStringGetDatum(opername),
 								 0, 0, 0);
 
+	/*
+	 * In typical scenarios, most if not all of the operators found by the
+	 * catcache search will end up getting returned; and there can be
+	 * quite a few, for common operator names such as '=' or '+'.  To
+	 * reduce the time spent in palloc, we allocate the result space as an
+	 * array large enough to hold all the operators.  The original coding
+	 * of this routine did a separate palloc for each operator, but
+	 * profiling revealed that the pallocs used an unreasonably large
+	 * fraction of parsing time.
+	 */
+#define SPACE_PER_OP MAXALIGN(sizeof(struct _FuncCandidateList) + sizeof(Oid))
+
+	if (catlist->n_members > 0)
+		resultSpace = palloc(catlist->n_members * SPACE_PER_OP);
+
 	for (i = 0; i < catlist->n_members; i++)
 	{
 		HeapTuple	opertup = &catlist->members[i]->tuple;
@@ -744,20 +732,16 @@ OpernameGetCandidates(List *names, char oprkind)
 		}
 		else
 		{
-			/*
-			 * Consider only opers that are in the search path and are not
-			 * in the temp namespace.
-			 */
-			List	   *nsp;
+			/* Consider only opers that are in the search path */
+			ListCell   *nsp;
 
 			foreach(nsp, namespaceSearchPath)
 			{
-				if (operform->oprnamespace == lfirsto(nsp) &&
-					operform->oprnamespace != myTempNamespace)
+				if (operform->oprnamespace == lfirst_oid(nsp))
 					break;
 				pathpos++;
 			}
-			if (nsp == NIL)
+			if (nsp == NULL)
 				continue;		/* oper is not in search path */
 
 			/*
@@ -811,8 +795,9 @@ OpernameGetCandidates(List *names, char oprkind)
 		/*
 		 * Okay to add it to result list
 		 */
-		newResult = (FuncCandidateList)
-			palloc(sizeof(struct _FuncCandidateList) + sizeof(Oid));
+		newResult = (FuncCandidateList) (resultSpace + nextResult);
+		nextResult += SPACE_PER_OP;
+
 		newResult->pathpos = pathpos;
 		newResult->oid = HeapTupleGetOid(opertup);
 		newResult->nargs = 2;
@@ -853,11 +838,11 @@ OperatorIsVisible(Oid oprid)
 	/*
 	 * Quick check: if it ain't in the path at all, it ain't visible.
 	 * Items in the system namespace are surely in the path and so we
-	 * needn't even do oidMember() for them.
+	 * needn't even do list_member_oid() for them.
 	 */
 	oprnamespace = oprform->oprnamespace;
 	if (oprnamespace != PG_CATALOG_NAMESPACE &&
-		!oidMember(oprnamespace, namespaceSearchPath))
+		!list_member_oid(namespaceSearchPath, oprnamespace))
 		visible = false;
 	else
 	{
@@ -873,7 +858,7 @@ OperatorIsVisible(Oid oprid)
 
 		visible = false;
 
-		clist = OpernameGetCandidates(makeList1(makeString(oprname)),
+		clist = OpernameGetCandidates(list_make1(makeString(oprname)),
 									  oprform->oprkind);
 
 		for (; clist; clist = clist->next)
@@ -924,17 +909,16 @@ OpclassGetCandidates(Oid amid)
 		Form_pg_opclass opcform = (Form_pg_opclass) GETSTRUCT(opctup);
 		int			pathpos = 0;
 		OpclassCandidateList newResult;
-		List	   *nsp;
+		ListCell   *nsp;
 
 		/* Consider only opclasses that are in the search path */
 		foreach(nsp, namespaceSearchPath)
 		{
-			if (opcform->opcnamespace == lfirsto(nsp) &&
-				opcform->opcnamespace != myTempNamespace)
+			if (opcform->opcnamespace == lfirst_oid(nsp))
 				break;
 			pathpos++;
 		}
-		if (nsp == NIL)
+		if (nsp == NULL)
 			continue;			/* opclass is not in search path */
 
 		/*
@@ -1020,16 +1004,13 @@ Oid
 OpclassnameGetOpcid(Oid amid, const char *opcname)
 {
 	Oid			opcid;
-	List	   *lptr;
+	ListCell   *l;
 
 	recomputeNamespacePath();
 
-	foreach(lptr, namespaceSearchPath)
+	foreach(l, namespaceSearchPath)
 	{
-		Oid			namespaceId = lfirsto(lptr);
-
-		if (namespaceId == myTempNamespace)
-			continue;			/* do not look in temp namespace */
+		Oid			namespaceId = lfirst_oid(l);
 
 		opcid = GetSysCacheOid(CLAAMNAMENSP,
 							   ObjectIdGetDatum(amid),
@@ -1070,11 +1051,11 @@ OpclassIsVisible(Oid opcid)
 	/*
 	 * Quick check: if it ain't in the path at all, it ain't visible.
 	 * Items in the system namespace are surely in the path and so we
-	 * needn't even do oidMember() for them.
+	 * needn't even do list_member_oid() for them.
 	 */
 	opcnamespace = opcform->opcnamespace;
 	if (opcnamespace != PG_CATALOG_NAMESPACE &&
-		!oidMember(opcnamespace, namespaceSearchPath))
+		!list_member_oid(namespaceSearchPath, opcnamespace))
 		visible = false;
 	else
 	{
@@ -1105,16 +1086,13 @@ Oid
 ConversionGetConid(const char *conname)
 {
 	Oid			conid;
-	List	   *lptr;
+	ListCell   *l;
 
 	recomputeNamespacePath();
 
-	foreach(lptr, namespaceSearchPath)
+	foreach(l, namespaceSearchPath)
 	{
-		Oid			namespaceId = lfirsto(lptr);
-
-		if (namespaceId == myTempNamespace)
-			continue;			/* do not look in temp namespace */
+		Oid			namespaceId = lfirst_oid(l);
 
 		conid = GetSysCacheOid(CONNAMENSP,
 							   PointerGetDatum(conname),
@@ -1154,11 +1132,11 @@ ConversionIsVisible(Oid conid)
 	/*
 	 * Quick check: if it ain't in the path at all, it ain't visible.
 	 * Items in the system namespace are surely in the path and so we
-	 * needn't even do oidMember() for them.
+	 * needn't even do list_member_oid() for them.
 	 */
 	connamespace = conform->connamespace;
 	if (connamespace != PG_CATALOG_NAMESPACE &&
-		!oidMember(connamespace, namespaceSearchPath))
+		!list_member_oid(namespaceSearchPath, connamespace))
 		visible = false;
 	else
 	{
@@ -1194,17 +1172,17 @@ DeconstructQualifiedName(List *names,
 	char	   *schemaname = NULL;
 	char	   *objname = NULL;
 
-	switch (length(names))
+	switch (list_length(names))
 	{
 		case 1:
-			objname = strVal(lfirst(names));
+			objname = strVal(linitial(names));
 			break;
 		case 2:
-			schemaname = strVal(lfirst(names));
+			schemaname = strVal(linitial(names));
 			objname = strVal(lsecond(names));
 			break;
 		case 3:
-			catalogname = strVal(lfirst(names));
+			catalogname = strVal(linitial(names));
 			schemaname = strVal(lsecond(names));
 			objname = strVal(lthird(names));
 
@@ -1214,7 +1192,8 @@ DeconstructQualifiedName(List *names,
 			if (strcmp(catalogname, get_database_name(MyDatabaseId)) != 0)
 				ereport(ERROR,
 						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				errmsg("cross-database references are not implemented")));
+						 errmsg("cross-database references are not implemented: %s",
+								NameListToString(names))));
 			break;
 		default:
 			ereport(ERROR,
@@ -1241,19 +1220,6 @@ LookupExplicitNamespace(const char *nspname)
 	Oid			namespaceId;
 	AclResult	aclresult;
 
-	/* check for pg_temp alias */
-	if (strcmp(nspname, "pg_temp") == 0)
-	{
-		if (OidIsValid(myTempNamespace))
-			return myTempNamespace;
-		/*
-		 * Since this is used only for looking up existing objects, there
-		 * is no point in trying to initialize the temp namespace here;
-		 * and doing so might create problems for some callers.
-		 * Just fall through and give the "does not exist" error.
-		 */
-	}
-
 	namespaceId = GetSysCacheOid(NAMESPACENAME,
 								 CStringGetDatum(nspname),
 								 0, 0, 0);
@@ -1276,28 +1242,21 @@ LookupExplicitNamespace(const char *nspname)
  *		format), determine what namespace the object should be created in.
  *		Also extract and return the object name (last component of list).
  *
- * Note: calling this may result in a CommandCounterIncrement operation,
- * if we have to create or clean out the temp namespace.
+ * This is *not* used for tables.  Hence, the TEMP table namespace is
+ * never selected as the creation target.
  */
 Oid
 QualifiedNameGetCreationNamespace(List *names, char **objname_p)
 {
 	char	   *schemaname;
+	char	   *objname;
 	Oid			namespaceId;
 
 	/* deconstruct the name list */
-	DeconstructQualifiedName(names, &schemaname, objname_p);
+	DeconstructQualifiedName(names, &schemaname, &objname);
 
 	if (schemaname)
 	{
-		/* check for pg_temp alias */
-		if (strcmp(schemaname, "pg_temp") == 0)
-		{
-			/* Initialize temp namespace if first time through */
-			if (!OidIsValid(myTempNamespace))
-				InitTempTableNamespace();
-			return myTempNamespace;
-		}
 		/* use exact schema given */
 		namespaceId = GetSysCacheOid(NAMESPACENAME,
 									 CStringGetDatum(schemaname),
@@ -1312,12 +1271,6 @@ QualifiedNameGetCreationNamespace(List *names, char **objname_p)
 	{
 		/* use the default creation namespace */
 		recomputeNamespacePath();
-		if (tempCreationPending)
-		{
-			/* Need to initialize temp namespace */
-			InitTempTableNamespace();
-			return myTempNamespace;
-		}
 		namespaceId = defaultCreationNamespace;
 		if (!OidIsValid(namespaceId))
 			ereport(ERROR,
@@ -1327,6 +1280,7 @@ QualifiedNameGetCreationNamespace(List *names, char **objname_p)
 
 	/* Note: callers will check for CREATE rights when appropriate */
 
+	*objname_p = objname;
 	return namespaceId;
 }
 
@@ -1339,17 +1293,17 @@ makeRangeVarFromNameList(List *names)
 {
 	RangeVar   *rel = makeRangeVar(NULL, NULL);
 
-	switch (length(names))
+	switch (list_length(names))
 	{
 		case 1:
-			rel->relname = strVal(lfirst(names));
+			rel->relname = strVal(linitial(names));
 			break;
 		case 2:
-			rel->schemaname = strVal(lfirst(names));
+			rel->schemaname = strVal(linitial(names));
 			rel->relname = strVal(lsecond(names));
 			break;
 		case 3:
-			rel->catalogname = strVal(lfirst(names));
+			rel->catalogname = strVal(linitial(names));
 			rel->schemaname = strVal(lsecond(names));
 			rel->relname = strVal(lthird(names));
 			break;
@@ -1375,13 +1329,13 @@ char *
 NameListToString(List *names)
 {
 	StringInfoData string;
-	List	   *l;
+	ListCell   *l;
 
 	initStringInfo(&string);
 
 	foreach(l, names)
 	{
-		if (l != names)
+		if (l != list_head(names))
 			appendStringInfoChar(&string, '.');
 		appendStringInfoString(&string, strVal(lfirst(l)));
 	}
@@ -1400,13 +1354,13 @@ char *
 NameListToQuotedString(List *names)
 {
 	StringInfoData string;
-	List	   *l;
+	ListCell   *l;
 
 	initStringInfo(&string);
 
 	foreach(l, names)
 	{
-		if (l != names)
+		if (l != list_head(names))
 			appendStringInfoChar(&string, '.');
 		appendStringInfoString(&string, quote_identifier(strVal(lfirst(l))));
 	}
@@ -1487,7 +1441,7 @@ FindConversionByName(List *name)
 	char	   *conversion_name;
 	Oid			namespaceId;
 	Oid			conoid;
-	List	   *lptr;
+	ListCell   *l;
 
 	/* deconstruct the name list */
 	DeconstructQualifiedName(name, &schemaname, &conversion_name);
@@ -1503,13 +1457,9 @@ FindConversionByName(List *name)
 		/* search for it in search path */
 		recomputeNamespacePath();
 
-		foreach(lptr, namespaceSearchPath)
+		foreach(l, namespaceSearchPath)
 		{
-			namespaceId = lfirsto(lptr);
-
-			if (namespaceId == myTempNamespace)
-				continue;			/* do not look in temp namespace */
-
+			namespaceId = lfirst_oid(l);
 			conoid = FindConversion(conversion_name, namespaceId);
 			if (OidIsValid(conoid))
 				return conoid;
@@ -1527,16 +1477,13 @@ Oid
 FindDefaultConversionProc(int4 for_encoding, int4 to_encoding)
 {
 	Oid			proc;
-	List	   *lptr;
+	ListCell   *l;
 
 	recomputeNamespacePath();
 
-	foreach(lptr, namespaceSearchPath)
+	foreach(l, namespaceSearchPath)
 	{
-		Oid			namespaceId = lfirsto(lptr);
-
-		if (namespaceId == myTempNamespace)
-			continue;			/* do not look in temp namespace */
+		Oid			namespaceId = lfirst_oid(l);
 
 		proc = FindDefaultConversion(namespaceId, for_encoding, to_encoding);
 		if (OidIsValid(proc))
@@ -1558,8 +1505,7 @@ recomputeNamespacePath(void)
 	List	   *namelist;
 	List	   *oidlist;
 	List	   *newpath;
-	List	   *l;
-	bool		temp_missing;
+	ListCell   *l;
 	Oid			firstNS;
 	MemoryContext oldcxt;
 
@@ -1587,7 +1533,6 @@ recomputeNamespacePath(void)
 	 * has already been accepted.)	Don't make duplicate entries, either.
 	 */
 	oidlist = NIL;
-	temp_missing = false;
 	foreach(l, namelist)
 	{
 		char	   *curname = (char *) lfirst(l);
@@ -1611,25 +1556,10 @@ recomputeNamespacePath(void)
 											 0, 0, 0);
 				ReleaseSysCache(tuple);
 				if (OidIsValid(namespaceId) &&
-					!oidMember(namespaceId, oidlist) &&
+					!list_member_oid(oidlist, namespaceId) &&
 					pg_namespace_aclcheck(namespaceId, userId,
 										  ACL_USAGE) == ACLCHECK_OK)
-					oidlist = lappendo(oidlist, namespaceId);
-			}
-		}
-		else if (strcmp(curname, "pg_temp") == 0)
-		{
-			/* pg_temp --- substitute temp namespace, if any */
-			if (OidIsValid(myTempNamespace))
-			{
-				if (!oidMember(myTempNamespace, oidlist))
-					oidlist = lappendo(oidlist, myTempNamespace);
-			}
-			else
-			{
-				/* If it ought to be the creation namespace, set flag */
-				if (oidlist == NIL)
-					temp_missing = true;
+					oidlist = lappend_oid(oidlist, namespaceId);
 			}
 		}
 		else
@@ -1639,49 +1569,47 @@ recomputeNamespacePath(void)
 										 CStringGetDatum(curname),
 										 0, 0, 0);
 			if (OidIsValid(namespaceId) &&
-				!oidMember(namespaceId, oidlist) &&
+				!list_member_oid(oidlist, namespaceId) &&
 				pg_namespace_aclcheck(namespaceId, userId,
 									  ACL_USAGE) == ACLCHECK_OK)
-				oidlist = lappendo(oidlist, namespaceId);
+				oidlist = lappend_oid(oidlist, namespaceId);
 		}
 	}
 
 	/*
-	 * Remember the first member of the explicit list.  (Note: this is
-	 * nominally wrong if temp_missing, but we need it anyway to distinguish
-	 * explicit from implicit mention of pg_catalog.)
+	 * Remember the first member of the explicit list.
 	 */
 	if (oidlist == NIL)
 		firstNS = InvalidOid;
 	else
-		firstNS = lfirsto(oidlist);
+		firstNS = linitial_oid(oidlist);
 
 	/*
 	 * Add any implicitly-searched namespaces to the list.	Note these go
 	 * on the front, not the back; also notice that we do not check USAGE
 	 * permissions for these.
 	 */
-	if (!oidMember(PG_CATALOG_NAMESPACE, oidlist))
-		oidlist = lconso(PG_CATALOG_NAMESPACE, oidlist);
+	if (!list_member_oid(oidlist, PG_CATALOG_NAMESPACE))
+		oidlist = lcons_oid(PG_CATALOG_NAMESPACE, oidlist);
 
 	if (OidIsValid(myTempNamespace) &&
-		!oidMember(myTempNamespace, oidlist))
-		oidlist = lconso(myTempNamespace, oidlist);
+		!list_member_oid(oidlist, myTempNamespace))
+		oidlist = lcons_oid(myTempNamespace, oidlist);
 
 	if (OidIsValid(mySpecialNamespace) &&
-		!oidMember(mySpecialNamespace, oidlist))
-		oidlist = lconso(mySpecialNamespace, oidlist);
+		!list_member_oid(oidlist, mySpecialNamespace))
+		oidlist = lcons_oid(mySpecialNamespace, oidlist);
 
 	/*
 	 * Now that we've successfully built the new list of namespace OIDs,
 	 * save it in permanent storage.
 	 */
 	oldcxt = MemoryContextSwitchTo(TopMemoryContext);
-	newpath = listCopy(oidlist);
+	newpath = list_copy(oidlist);
 	MemoryContextSwitchTo(oldcxt);
 
 	/* Now safe to assign to state variable. */
-	freeList(namespaceSearchPath);
+	list_free(namespaceSearchPath);
 	namespaceSearchPath = newpath;
 
 	/*
@@ -1689,16 +1617,9 @@ recomputeNamespacePath(void)
 	 */
 	firstExplicitNamespace = firstNS;
 	if (OidIsValid(mySpecialNamespace))
-	{
 		defaultCreationNamespace = mySpecialNamespace;
-		/* don't have to create temp in this state */
-		tempCreationPending = false;
-	}
 	else
-	{
 		defaultCreationNamespace = firstNS;
-		tempCreationPending = temp_missing;
-	}
 
 	/* Mark the path valid. */
 	namespaceSearchPathValid = true;
@@ -1706,8 +1627,8 @@ recomputeNamespacePath(void)
 
 	/* Clean up. */
 	pfree(rawname);
-	freeList(namelist);
-	freeList(oidlist);
+	list_free(namelist);
+	list_free(oidlist);
 }
 
 /*
@@ -1720,18 +1641,18 @@ InitTempTableNamespace(void)
 	char		namespaceName[NAMEDATALEN];
 	Oid			namespaceId;
 
-	Assert(!OidIsValid(myTempNamespace));
-
 	/*
 	 * First, do permission check to see if we are authorized to make temp
 	 * tables.	We use a nonstandard error message here since
 	 * "databasename: permission denied" might be a tad cryptic.
 	 *
-	 * Note we apply the check to the session user, not the currently active
-	 * userid, since we are not going to change our minds about temp table
-	 * availability during the session.
+	 * Note that ACL_CREATE_TEMP rights are rechecked in
+	 * pg_namespace_aclmask; that's necessary since current user ID could
+	 * change during the session. But there's no need to make the
+	 * namespace in the first place until a temp table creation request is
+	 * made by someone with appropriate rights.
 	 */
-	if (pg_database_aclcheck(MyDatabaseId, GetSessionUserId(),
+	if (pg_database_aclcheck(MyDatabaseId, GetUserId(),
 							 ACL_CREATE_TEMP) != ACLCHECK_OK)
 		ereport(ERROR,
 				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
@@ -1774,7 +1695,9 @@ InitTempTableNamespace(void)
 	 */
 	myTempNamespace = namespaceId;
 
-	firstTempTransaction = true;
+	/* It should not be done already. */
+	AssertState(myTempNamespaceSubID == InvalidSubTransactionId);
+	myTempNamespaceSubID = GetCurrentSubTransactionId();
 
 	namespaceSearchPathValid = false;	/* need to rebuild list */
 }
@@ -1793,7 +1716,7 @@ AtEOXact_Namespace(bool isCommit)
 	 * temp tables at backend shutdown.  (We only want to register the
 	 * callback once per session, so this is a good place to do it.)
 	 */
-	if (firstTempTransaction)
+	if (myTempNamespaceSubID != InvalidSubTransactionId)
 	{
 		if (isCommit)
 			on_shmem_exit(RemoveTempRelationsCallback, 0);
@@ -1802,7 +1725,7 @@ AtEOXact_Namespace(bool isCommit)
 			myTempNamespace = InvalidOid;
 			namespaceSearchPathValid = false;	/* need to rebuild list */
 		}
-		firstTempTransaction = false;
+		myTempNamespaceSubID = InvalidSubTransactionId;
 	}
 
 	/*
@@ -1815,6 +1738,32 @@ AtEOXact_Namespace(bool isCommit)
 	}
 }
 
+/*
+ * AtEOSubXact_Namespace
+ *
+ * At subtransaction commit, propagate the temp-namespace-creation
+ * flag to the parent subtransaction.
+ *
+ * At subtransaction abort, forget the flag if we set it up.
+ */
+void
+AtEOSubXact_Namespace(bool isCommit, SubTransactionId mySubid,
+					  SubTransactionId parentSubid)
+{
+	if (myTempNamespaceSubID == mySubid)
+	{
+		if (isCommit)
+			myTempNamespaceSubID = parentSubid;
+		else
+		{
+			myTempNamespaceSubID = InvalidSubTransactionId;
+			/* TEMP namespace creation failed, so reset state */
+			myTempNamespace = InvalidOid;
+			namespaceSearchPathValid = false;	/* need to rebuild list */
+		}
+	}
+}
+
 /*
  * Remove all relations in the specified temp namespace.
  *
@@ -1845,7 +1794,7 @@ RemoveTempRelations(Oid tempNamespaceId)
  * Callback to remove temp relations at backend exit.
  */
 static void
-RemoveTempRelationsCallback(void)
+RemoveTempRelationsCallback(int code, Datum arg)
 {
 	if (OidIsValid(myTempNamespace))	/* should always be true */
 	{
@@ -1866,11 +1815,11 @@ RemoveTempRelationsCallback(void)
 
 /* assign_hook: validate new search_path, do extra actions as needed */
 const char *
-assign_search_path(const char *newval, bool doit, bool interactive)
+assign_search_path(const char *newval, bool doit, GucSource source)
 {
 	char	   *rawname;
 	List	   *namelist;
-	List	   *l;
+	ListCell   *l;
 
 	/* Need a modifiable copy of string */
 	rawname = pstrdup(newval);
@@ -1880,7 +1829,7 @@ assign_search_path(const char *newval, bool doit, bool interactive)
 	{
 		/* syntax error in name list */
 		pfree(rawname);
-		freeList(namelist);
+		list_free(namelist);
 		return NULL;
 	}
 
@@ -1888,13 +1837,20 @@ assign_search_path(const char *newval, bool doit, bool interactive)
 	 * If we aren't inside a transaction, we cannot do database access so
 	 * cannot verify the individual names.	Must accept the list on faith.
 	 */
-	if (interactive && IsTransactionState())
+	if (source >= PGC_S_INTERACTIVE && IsTransactionState())
 	{
 		/*
 		 * Verify that all the names are either valid namespace names or
-		 * "$user" or "pg_temp".  We do not require $user to correspond to a
-		 * valid namespace, and pg_temp might not exist yet.  We do not check
-		 * for USAGE rights, either; should we?
+		 * "$user".  We do not require $user to correspond to a valid
+		 * namespace.  We do not check for USAGE rights, either; should
+		 * we?
+		 *
+		 * When source == PGC_S_TEST, we are checking the argument of an
+		 * ALTER DATABASE SET or ALTER USER SET command.  It could be that
+		 * the intended use of the search path is for some other database,
+		 * so we should not error out if it mentions schemas not present
+		 * in the current database.  We reduce the message to NOTICE
+		 * instead.
 		 */
 		foreach(l, namelist)
 		{
@@ -1902,19 +1858,17 @@ assign_search_path(const char *newval, bool doit, bool interactive)
 
 			if (strcmp(curname, "$user") == 0)
 				continue;
-			if (strcmp(curname, "pg_temp") == 0)
-				continue;
 			if (!SearchSysCacheExists(NAMESPACENAME,
 									  CStringGetDatum(curname),
 									  0, 0, 0))
-				ereport(ERROR,
+				ereport((source == PGC_S_TEST) ? NOTICE : ERROR,
 						(errcode(ERRCODE_UNDEFINED_SCHEMA),
 					   errmsg("schema \"%s\" does not exist", curname)));
 		}
 	}
 
 	pfree(rawname);
-	freeList(namelist);
+	list_free(namelist);
 
 	/*
 	 * We mark the path as needing recomputation, but don't do anything
@@ -1945,11 +1899,10 @@ InitializeSearchPath(void)
 		MemoryContext oldcxt;
 
 		oldcxt = MemoryContextSwitchTo(TopMemoryContext);
-		namespaceSearchPath = makeListo1(PG_CATALOG_NAMESPACE);
+		namespaceSearchPath = list_make1_oid(PG_CATALOG_NAMESPACE);
 		MemoryContextSwitchTo(oldcxt);
 		defaultCreationNamespace = PG_CATALOG_NAMESPACE;
 		firstExplicitNamespace = PG_CATALOG_NAMESPACE;
-		tempCreationPending = false;
 		namespaceSearchPathValid = true;
 		namespaceUser = GetUserId();
 	}
@@ -1979,15 +1932,12 @@ NamespaceCallback(Datum arg, Oid relid)
 }
 
 /*
- * Fetch the active search path, expressed as a List of OIDs.
+ * Fetch the active search path. The return value is a palloc'ed list
+ * of OIDs; the caller is responsible for freeing this storage as
+ * appropriate.
  *
  * The returned list includes the implicitly-prepended namespaces only if
  * includeImplicit is true.
- *
- * NB: caller must treat the list as read-only!
- *
- * Note: calling this may result in a CommandCounterIncrement operation,
- * if we have to create or clean out the temp namespace.
  */
 List *
 fetch_search_path(bool includeImplicit)
@@ -1996,24 +1946,11 @@ fetch_search_path(bool includeImplicit)
 
 	recomputeNamespacePath();
 
-	/*
-	 * If the temp namespace should be first, force it to exist.  This is
-	 * so that callers can trust the result to reflect the actual default
-	 * creation namespace.  It's a bit bogus to do this here, since
-	 * current_schema() is supposedly a stable function without side-effects,
-	 * but the alternatives seem worse.
-	 */
-	if (tempCreationPending)
-	{
-		InitTempTableNamespace();
-		recomputeNamespacePath();
-	}
-
-	result = namespaceSearchPath;
+	result = list_copy(namespaceSearchPath);
 	if (!includeImplicit)
 	{
-		while (result && lfirsto(result) != firstExplicitNamespace)
-			result = lnext(result);
+		while (result && linitial_oid(result) != firstExplicitNamespace)
+			result = list_delete_first(result);
 	}
 
 	return result;
